from optparse import OptionParser
import sys, os
from utilities.align_parsers import blat, exonerate #Needs the exonerate and blat parser libraries
from utilities import intspan #Use custom intspan library

#Class handles candidate fusion alignment pairs from a single contig
class Fusion:

    #Initialize our variables
    #contig1 in pair
    align1 = None
    #contig2 in pair
    align2 = None
    #contig number
    contig = None
    #bp overlap in contig alignments
    overlap_query = None
    #fraction of total length that has contig alignment overlap
    overlap_query_fraction = None
    #bp overlap in target alignments
    overlap_target = None
    #fraction of total length that has target alignment overlap
    overlap_target_fraction = None
    #bp's covered by the two alignments
    contig_cover = None
    #fraction of total contig covered by two alingments
    contig_cover_fraction = None

    def __init__(self, a1, a2):
        #Read in the alignment pairs
        self.align1 = a1
        self.align2 = a2
        #attributes of the contig 
        self.contig = a1.query
        self.contig_size = a1.query_len

    def details(self):
        #function formats the output lines from the alignment pair
        ctg_fields = self.contig.split(" ")
        ctg = "%s(%sbp)" % (ctg_fields[0], self.contig_size)
        target = "%s:%s-%s,%s:%s-%s" % (self.align1.target, self.align1.tstart, self.align1.tend, self.align2.target, self.align2.tstart, self.align2.tend)
        query = "%s-%s,%s-%s" % (self.align1.qstart, self.align1.qend, self.align2.qstart, self.align2.qend)

        align_len1 = float(self.align1.qend) - float(self.align1.qstart) + 1
        align_len2 = float(self.align2.qend) - float(self.align2.qstart) + 1
        align_frac1 = align_len1/float(self.align1.query_len)
        align_frac2 = align_len2/float(self.align2.query_len)

        params = "TO:%.2f,CO:%.2f,CC:%.2f,I1:%.1f,I2:%.1f,AF1:%.2f,AF2:%.2f" % (self.overlap_target_fraction, self.overlap_query_fraction, self.contig_cover_fraction, float(self.align1.identity), float(self.align2.identity), align_frac1, align_frac2)
        
        return "CTG:%s TARGET:%s CONTIG:%s %s" % (ctg, target, query, params)

#Class indentifies candidate fusion contigs (contigs that have split alignments to distinct genomic regions)
class FusionFinder:

    params = None
    fusions = None
    debug = None
    
    def __init__(self, params, debug=False):
        self.params = params
        self.fusions = []
        self.debug = debug

        #debug optipon
        if debug:
            print self.params

    #identifies alignment pairs exhibiting characteristics expected in a true fusion
    def get_split_aligns(self, aligns):

        if self.debug:
            for i in range(len(aligns)):
                print i,aligns[i].score,aligns[i].identity,aligns[i].match_len,aligns[i].query,aligns[i].qstart,aligns[i].qend,aligns[i].target,aligns[i].tstart,aligns[i].tend
            print "---"
        
        skip = None
        #iterate over the alignment pair combos (top 5 alignments currently kept)
        for i in range(len(aligns)):
            #if the skip flag was set
            if skip and aligns[i].query == aligns[i-1].query:
                continue
            else:
                skip = False

            #check if any member of the alignment set has a single alignment over 95% of the total length; if so, skip
            if (float(aligns[i].qend) - float(aligns[i].qstart)) / float(aligns[i].query_len) >= self.params['max_match_fraction']:
                if self.debug:
                    print "skip ",aligns[i].query
                skip = True
                continue
            
            #Check the partner alignment
            for j in range(i+1,len(aligns)):
                #if we're onto the next contig
                if not aligns[j].query == aligns[i].query:
                    break

                #check if any member of the alignment set has a single alignment over 95% of the total length; if so, skip
                if (float(aligns[j].qend) - float(aligns[j].qstart)) / float(aligns[j].query_len) >= self.params['max_match_fraction']:
                    if self.debug:
                        print "skip ",aligns[j].query
                    skip = True
                    break

            if skip:
                continue

            for j in range(i+1,len(aligns)):
                if not aligns[j].query == aligns[i].query:
                    break
                
                if aligns[i].query == aligns[j].query and aligns[i].target == aligns[j].target and (aligns[i].tstart == aligns[j].tstart or aligns[i].tend == aligns[j].tend):
                    continue

                #if one of the alignments is a subset of the other
                if intspan.subsume([aligns[i].qstart, aligns[i].qend], [aligns[j].qstart, aligns[j].qend]) or intspan.subsume([aligns[j].qstart, aligns[j].qend], [aligns[i].qstart, aligns[i].qend]):
                    continue

                #Check how much the query alignments between the contigs overlap
                (overlap_query, overlap_query_fraction) = self.overlap([aligns[i].qstart, aligns[i].qend], [aligns[j].qstart, aligns[j].qend], aligns[i].query_len)

                #Check if the overlap query between the contigs is too high
                if not overlap_query or overlap_query_fraction < self.params['max_query_olap']:
                    #Get the target query overlap
                    (overlap_target, overlap_target_fraction) = self.overlap([aligns[i].tstart, aligns[i].tend], [aligns[j].tstart, aligns[j].tend], aligns[i].query_len)

                    #if targets are the same, means no overlap
                    if aligns[i].target == aligns[j].target:
                        (overlap_target, overlap_target_fraction) = self.overlap([aligns[i].tstart, aligns[i].tend], [aligns[j].tstart, aligns[j].tend], aligns[i].query_len)
                    else:
                        (overlap_target, overlap_target_fraction) = (0.0, 0.0)

                    #Check if the target overlap between the contigs is too high
                    if overlap_target_fraction > self.params['max_target_olap']:
                        continue

                    #Get the coverage of both alignments
                    (contig_cover, contig_cover_fraction) = self.combined_aligned(aligns[i], aligns[j], overlap_query)

                    #Check if the cover enough of the contig between the two alignments is too low
                    if contig_cover_fraction > self.params['min_fusion_fraction']:
                        if self.debug:
                            print "good",i,j,aligns[i].query_len,aligns[i].match_len,float(aligns[i].match_len) / float(aligns[i].query_len),aligns[i].query,aligns[i].score,aligns[i].identity,aligns[i].target,aligns[i].tstart,aligns[i].tend,aligns[j].query,aligns[j].score,aligns[j].identity,aligns[j].target,aligns[j].tstart,aligns[j].tend,overlap_target,overlap_target_fraction,aligns[i].qstart,aligns[i].qend,aligns[j].qstart,aligns[j].qend,overlap_query,overlap_query_fraction,contig_cover_fraction

                        #Create a fusion object as we've passed all the filters
                        fusion = Fusion(aligns[i], aligns[j])
                        fusion.overlap_query_fraction = overlap_query_fraction
                        fusion.overlap_target_fraction = overlap_target_fraction
                        fusion.contig_cover_fraction = contig_cover_fraction
                        self.fusions.append(fusion)
                    else:
                        if self.debug:
                             print "failed:",i,j,aligns[i].query_len,aligns[i].match_len,float(aligns[i].match_len) / float(aligns[i].query_len),aligns[i].query,aligns[i].score,aligns[i].identity,aligns[i].target,aligns[i].tstart,aligns[i].tend,aligns[j].query,aligns[j].score,aligns[j].identity,aligns[j].target,aligns[j].tstart,aligns[j].tend,overlap_target,overlap_target_fraction,aligns[i].qstart,aligns[i].qend,aligns[j].qstart,aligns[j].qend,overlap_query,overlap_query_fraction,contig_cover_fraction

    #Function checks the overlap between two ranges and gives a proportion of the total length covered
    def overlap(self, span1, span2, total_len):
        s1 = [int(span1[0]), int(span1[1])]
        s2 = [int(span2[0]), int(span2[1])]
        s1.sort(lambda x,y:x-y)
        s2.sort(lambda x,y:x-y)

        olap = intspan.overlap(s1, s2)
        olap_fraction = 0.0
        if olap:
            olap_fraction = float(olap) / float(total_len)

        return (olap, olap_fraction)

    #Functions gives the total bases covered by both alignments and the portion of the total contig length covered
    def combined_aligned(self, align1, align2, olap):
        align1_aligned =  int(align1.qend) - int(align1.qstart) + 1
        align2_aligned =  int(align2.qend) - int(align2.qstart) + 1
        total_aligned = align1_aligned + align2_aligned - olap
        aligned_fraction = float(total_aligned) / float(align1.query_len)

        return (total_aligned, aligned_fraction)
        
    #reads in the alignment file and assigns the appropriate parser
    def find(self, file):
        #find best alignments
        aligns = []
        filters = {'count':self.params['num_aligns'], 'identity':self.params['min_identity']}
        ext = os.path.splitext(file)[1]
        if ext == ".psl":
            aligns = blat.parse(file, filters, noblocks=True, noline=True)
        else:
            aligns = exonerate.parse(file, filters)
            
        sys.stderr.write("done parsing alignments\n")
        
        #identify splitpairs
        self.get_split_aligns(aligns)


    #Generate the output file
    def output(self, out_file, append):
        if append:
            out = open(out_file, 'a')
        else:
            out = open(out_file, 'w')

        for fusion in self.fusions:
            out.write(fusion.details() + "\n")

        out.close()
        

#Parse the arguments
if __name__ == '__main__':
    usage = "Usage: %proj alignment_file output_file"

    parser = OptionParser(usage=usage)
    parser.add_option("-m", "--max_match_fraction", dest="match_fraction", help="maximum match fraction for individual alignment. Default:0.95", default=0.95, type='float')
    parser.add_option("-q", "--max_query_overlap", dest="query_olap", help="maximum query overlap. Default:0.1", default=0.1, type='float')
    parser.add_option("-t", "--max_target_overlap", dest="target_olap", help="maximum target overlap. Default:0.0", default=0.0, type='float')
    parser.add_option("-f", "--min_fusion_fraction", dest="fusion_fraction", help="minimum fusion fraction. Default:0.85", default=0.85, type='float')
    parser.add_option("-a", "--number_aligments", dest="num_aligns", help="number alignments. Default:5", default=5, type='int')
    parser.add_option("-i", "--min_identity", dest="min_identity", help="minimum identity. Default:90", type='float', default=90.0)
    parser.add_option("-d", "--debug", dest="debug", help="debug", action="store_true", default=False)
    parser.add_option("-e", "--append", dest="append", help="append to output", action="store_true", default=False)
    
    (options, args) = parser.parse_args()

    params = {}
    params['max_match_fraction'] = options.match_fraction
    params['max_query_olap'] = options.query_olap
    params['max_target_olap'] = options.target_olap
    params['min_fusion_fraction'] = options.fusion_fraction
    params['min_identity'] = options.min_identity
    params['num_aligns'] = options.num_aligns
    
    if len(args) == 2:
        ff = FusionFinder(params, debug=options.debug)
        ff.find(args[0])

        ff.output(args[1], append=options.append)

